change_spec = {
    "path.to.old_api": {
        "alias": [
            "path.to.old_api_alias1",
            "path.to1.to2.old_api_alias2",
            ],
        "update_to": "path.to.new_api",
        "warning": "this api is deprecated.",
        "args_list": ["arg1", "arg2"],
        "args_change": [
            ("arg2", "arg2_rename"),
            ("arg3", ""),
            ("", "new_arg", "default_value"),
            ],
        "args_warning": {"arg1":"warning message"},
        "args_transformer": "_default_transformer",
    },
    "paddle.optimizer.AdamOptimizer": {
        "alias": [
            "paddle.fluid.optimizer.AdamOptimizer",
            ],
        "update_to": "paddle.AdamOptimizer",
    },
    "paddle.optimizer.TestOptimizer": {
        "warning": "this api is deprecated, use paddle.TestOptimizer please.",
    },

}



import sys
import logging


__all__ = [
        'logger',
        'log_to_file',
        ]

class ColorFormatter(logging.Formatter):
    """Logging Formatter to add colors and count warning / errors"""
    def __init__(self, fmt=None, datefmt=None, style='%'):
        super().__init__(fmt, datefmt, style)

        grey = "\x1b[38;21m"
        yellow = "\x1b[33;21m"
        red = "\x1b[31;21m"
        bold_red = "\x1b[31;1m"
        reset = "\x1b[0m"

        self.FORMATS = {
            logging.DEBUG: grey + fmt + reset,
            logging.INFO: grey + fmt + reset,
            logging.WARNING: yellow + fmt + reset,
            logging.ERROR: red + fmt + reset,
            logging.CRITICAL: bold_red + fmt + reset
        }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)

def log_to_file(log_filepath="report.log"):
    log_filepath = log_filepath or "report.log"
    file_handler = logging.FileHandler(log_filepath)
    log_format = logging.Formatter('%(asctime)s - %(filename)s:%(lineno)d - %(levelname)s - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    file_handler.setFormatter(log_format)
    logger.addHandler(file_handler)

def _build_default_logger():
    logger = logging.getLogger('paddle1to2')
    logger.setLevel("INFO")
    
    console_handler = logging.StreamHandler(stream=sys.stdout) # default stream is sys.stderr
    log_format = ColorFormatter('%(asctime)s - %(filename)s:%(lineno)d - %(levelname)s - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    console_handler.setFormatter(log_format)
    logger.addHandler(console_handler)

    return logger

logger = _build_default_logger()



#!/usr/bin/env python3
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from pathlib import Path
from unittest import TestCase, mock

from ..query import Query
from ..tool import BadTransform, BowlerTool, log
from ..types import BowlerQuit

target = Path(__file__).parent / "smoke-target.py"
hunks = [
    [
        f"--- {target}",
        f"+++ {target}",
        "@@ -7,8 +7,8 @@",
        " ",
        " ",
        " # todo: i18n",
        '-print("Hello world!")',
        '+print(tr("Hello world!"))',
        " ",
        " ",
        "-def foo():",
        '+def foo(bar="something"):',
        "     pass",
    ],
    [
        f"--- {target}",
        f"+++ {target}",
        "@@ -10,11 +10,11 @@",
        " ",
        " ",
        " # todo: i18n",
        '-print("Hello world!")',
        '+print(tr("Hello world!"))',
        " ",
        " ",
        "-def foo():",
        '+def foo(bar="something"):',
        "     pass",
    ],
]


class ToolTest(TestCase):
    def setUp(self):
        self.orig_file = Path(__file__).parent / "smoke-target.py.orig"
        self.rej_file = Path(__file__).parent / "smoke-target.py.rej"
        echo_patcher = mock.patch("bowler.tool.click.echo")
        secho_patcher = mock.patch("bowler.tool.click.secho")
        self.addCleanup(echo_patcher.stop)
        self.addCleanup(secho_patcher.stop)
        self.mock_echo = echo_patcher.start()
        self.mock_secho = secho_patcher.start()
        self.addCleanup(self.cleanup_files)

    def cleanup_files(self):
        if os.path.isfile(self.orig_file):
            os.rename(self.orig_file, target)
        if os.path.isfile(self.rej_file):
            os.remove(self.rej_file)

    @mock.patch("bowler.tool.sh.patch")
    def test_process_hunks_patch_called_correctly(self, mock_patch):
        tool = BowlerTool(Query().compile(), write=True, interactive=False, silent=True)

        tool.process_hunks(target, hunks)
        string_hunks = ""
        for hunk in hunks:
            string_hunks += "\n".join(hunk[2:]) + "\n"
        string_hunks = f"--- {target}\n+++ {target}\n" + string_hunks
        mock_patch.assert_called_with("-u", target, _in=string_hunks.encode("utf-8"))

    @mock.patch.object(log, "exception")
    def test_process_hunks_invalid_hunks(self, mock_log):
        tool = BowlerTool(Query().compile(), write=True, interactive=False, silent=True)

        tool.process_hunks(target, hunks)
        mock_log.assert_called_with(
            f"hunks failed to apply, rejects saved to {target}.rej"
        )
        self.assertTrue(os.path.isfile(self.rej_file))
        self.assertTrue(os.path.isfile(self.orig_file))

    @mock.patch.object(log, "exception")
    def test_process_hunks_no_hunks(self, mock_log):
        tool = BowlerTool(Query().compile(), write=True, interactive=False)
        empty_hunks = [[]]
        tool.process_hunks(target, empty_hunks)
        patch_stderr = "/usr/bin/patch: **** Only garbage was found in the patch input."
        mock_log.assert_called_with(f"failed to apply patch hunk: {patch_stderr}")

    @mock.patch("bowler.tool.prompt_user")
    @mock.patch("bowler.tool.sh.patch")
    def test_process_hunks_after_skip_rest(self, mock_patch, mock_prompt):
        # Test that we apply the hunks that have been 'yessed' and nothing more
        tool = BowlerTool(Query().compile(), silent=False)
        mock_prompt.side_effect = ["y", "d"]
        tool.process_hunks(target, hunks)
        mock_patch.assert_called_once_with(
            "-u", target, _in="\n".join(hunks[0]).encode("utf-8") + b"\n"
        )

    @mock.patch("bowler.tool.prompt_user")
    @mock.patch("bowler.tool.sh.patch")
    def test_process_hunks_after_quit(self, mock_patch, mock_prompt):
        # Test that we apply the hunks that have been 'yessed' and nothing more
        tool = BowlerTool(Query().compile(), silent=False)
        mock_prompt.side_effect = ["y", "q"]
        with self.assertRaises(BowlerQuit):
            tool.process_hunks(target, hunks)
        mock_patch.assert_called_once_with(
            "-u", target, _in="\n".join(hunks[0]).encode("utf-8") + b"\n"
        )

    @mock.patch("bowler.tool.prompt_user")
    @mock.patch("bowler.tool.sh.patch")
    def test_process_hunks_after_auto_yes(self, mock_patch, mock_prompt):
        tool = BowlerTool(Query().compile(), silent=False)
        mock_prompt.side_effect = ["a"]
        tool.process_hunks(target, hunks)
        joined_hunks = "".join(["\n".join(hunk[2:]) + "\n" for hunk in hunks])
        encoded_hunks = f"--- {target}\n+++ {target}\n{joined_hunks}".encode("utf-8")
        mock_patch.assert_called_once_with("-u", target, _in=encoded_hunks)

    @mock.patch("bowler.tool.prompt_user")
    @mock.patch("bowler.tool.sh.patch")
    def test_process_hunks_after_no_then_yes(self, mock_patch, mock_prompt):
        tool = BowlerTool(Query().compile(), silent=False)
        mock_prompt.side_effect = ["n", "y"]
        tool.process_hunks(target, hunks)
        mock_patch.assert_called_once_with(
            "-u", target, _in="\n".join(hunks[1]).encode("utf-8") + b"\n"
        )

    @mock.patch("bowler.tool.prompt_user")
    @mock.patch("bowler.tool.sh.patch")
    def test_process_hunks_after_only_no(self, mock_patch, mock_prompt):
        tool = BowlerTool(Query().compile(), silent=False)
        mock_prompt.side_effect = ["n", "n"]
        tool.process_hunks(target, hunks)
        mock_patch.assert_not_called()

    def test_validate_print(self):
        tool = BowlerTool(Query().compile(), silent=False)
        # Passes
        tool.processed_file(
            new_text="print('str')", filename="foo.py", old_text="print('str')"
        )
        # Fails
        with self.assertRaises(BadTransform):
            tool.processed_file(
                new_text="print 'str'", filename="foo.py", old_text="print('str')"
            )

    def test_validate_completely_invalid(self):
        tool = BowlerTool(Query().compile(), silent=False)
        with self.assertRaises(BadTransform):
            tool.processed_file(new_text="x=1///2", filename="foo.py", old_text="x=1/2")



#!/usr/bin/env python3
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

if __name__ == "__main__":  # pragma: no cover
    unittest.main(module="bowler.tests", verbosity=2)



#!/usr/bin/env python3
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import sys
import unittest
from io import StringIO

from fissix import pygram, pytree
from fissix.pgen2.driver import Driver

from ..type_inference import OP_MIN_TYPE, InferredType, numeric_expr_type
from ..types import LN, SYMBOL, TOKEN
from .lib import BowlerTestCase

BINARY_OPERATORS = ["+", "-", "*", "**", "<<", ">>", "|", "&", "^", "%", "<"]

UNARY_OPERATORS = ["~", "-", "+"]
# TODO 'is', 'not'

SAMPLE_EXPRESSIONS = [
    # and/or
    ("True", InferredType.BOOL),
    ("True or False", InferredType.BOOL),
    ("True or 1", InferredType.INT),
    ("1 or 1.0", InferredType.FLOAT),
    # Calls
    ("bool(x)", InferredType.BOOL),
    ("int(x)", InferredType.INT),
    ("len(x)", InferredType.INT),
    ("float(x)", InferredType.FLOAT),
    # Basic
    ("func()", InferredType.INT_OR_FLOAT),
    ("x+1", InferredType.INT_OR_FLOAT),
    ("x*1", InferredType.INT_OR_FLOAT),
    ("1+x", InferredType.INT_OR_FLOAT),
    ("1*x", InferredType.INT_OR_FLOAT),
    # Single
    ("1+1", InferredType.INT),
    ("1.0+1.0", InferredType.FLOAT),
    # Mixed
    ("1+1.0", InferredType.FLOAT),
    ("1+x/2", InferredType.FLOAT),  # py3
    ("1.0+x/2", InferredType.FLOAT),
    # Division
    ("1/1", InferredType.FLOAT),
    ("1/2.0", InferredType.FLOAT),
    ("1.0/2", InferredType.FLOAT),
    ("1/x", InferredType.FLOAT),
    ("1.0/x", InferredType.FLOAT),
    ("True/True", InferredType.FLOAT),
    ("1j/2", InferredType.COMPLEX),
    # Floor Division
    ("1//2", InferredType.INT),
    ("1.0//2.0", InferredType.INT),
]

SAMPLE_PY2_EXPRESSIONS = [
    # Py2 Division
    ("1/1", InferredType.INT),
    ("1/2.0", InferredType.FLOAT),
    ("1.0/2", InferredType.FLOAT),
    ("1/x", InferredType.INT_OR_FLOAT),
    ("1.0/x", InferredType.FLOAT),
    ("True/True", InferredType.INT),
    ("1j/2", InferredType.COMPLEX),
]


def _produce_test(lcals, gen_func, args):
    t = gen_func(*args)
    t.__name__ += str(args)
    lcals[t.__name__] = t


def tree(input: str) -> LN:
    print(f"Input is {repr(input)}")
    driver = Driver(pygram.python_grammar_no_print_statement, convert=pytree.convert)
    return driver.parse_string(input)


def map_type(o):
    if isinstance(o, complex):
        return InferredType.COMPLEX
    elif isinstance(o, float):
        return InferredType.FLOAT
    elif isinstance(o, bool):
        return InferredType.BOOL
    elif isinstance(o, int):
        return InferredType.INT


class OpMinTypeTest(BowlerTestCase):
    """
    Verifies that the generated OP_MIN_TYPE matches that of the current Python
    interpreter, and that `numeric_expr_type` agrees.
    """

    def _run_test(self, expression_str, expected_type):
        t = tree(expression_str)
        expr = t.children[0].children[0]
        # t_op = expr.children[1].type
        # expr_type = pytree.type_repr(expr.type)
        # key_type = TOKEN.tok_name[t_op]

        inferred_result = numeric_expr_type(expr)

        self.assertEqual(expected_type, inferred_result, repr(expr))

    def gen_test_binop(op):
        def test_binop(self):
            for lhs in ("True", "1", "1.0", "2j"):
                for rhs in ("True", "1", "1.0", "2j"):
                    snippet = f"{lhs} {op} {rhs}\n"
                    try:
                        real_result = eval(snippet, {}, {})
                    except TypeError:
                        continue
                    self._run_test(snippet, map_type(real_result))

        return test_binop

    def gen_test_uniop(op):
        def test_uniop(self):
            for rhs in ("True", "1", "1.0", "2j"):
                snippet = f"{op} {rhs}\n"
                try:
                    real_result = eval(snippet, {}, {})
                except TypeError:
                    continue
                self._run_test(snippet, map_type(real_result))

        return test_uniop

    def gen_test_min_type(op):
        def test_min_type(self):
            snippet = f"True {op} True\n"
            real_result = eval(snippet, {}, {})

            t = tree(snippet)
            expr = t.children[0].children[0]
            t_op = expr.children[1].type
            expr_type = pytree.type_repr(expr.type)
            key_type = TOKEN.tok_name[t_op]

            self.assertEqual(map_type(real_result), OP_MIN_TYPE.get(t_op), key_type)

        return test_min_type

    def gen_test_min_type_unary(op):
        def test_min_type_unary(self):
            snippet = f"{op} True\n"
            real_result = eval(snippet, {}, {})

            t = tree(snippet)

            expr = t.children[0].children[0]
            t_op = expr.children[0].type
            expr_type = pytree.type_repr(expr.type)
            key_type = TOKEN.tok_name[t_op]

            self.assertEqual(map_type(real_result), OP_MIN_TYPE.get(t_op), expr_type)

        return test_min_type_unary

    # This produces real methods that can have normal decorators on them, with
    # a proper pass/fail count (unlike subTest).
    for i, op in enumerate(BINARY_OPERATORS):
        _produce_test(locals(), gen_test_binop, (op,))
        _produce_test(locals(), gen_test_min_type, (op,))

    for i, op in enumerate(UNARY_OPERATORS):
        _produce_test(locals(), gen_test_uniop, (op,))
        _produce_test(locals(), gen_test_min_type_unary, (op,))


class ExpressionTest(BowlerTestCase):
    def gen_test_expression(expression_str, expected_type, use_py2_division=False):
        def test_expression(self):
            expr = tree(expression_str).children[0].children[0]
            inferred_type = numeric_expr_type(
                expr, use_py2_division, type_for_unknown=InferredType.INT_OR_FLOAT
            )
            self.assertEqual(expected_type, inferred_type)

        return test_expression

    for expr, expected in SAMPLE_EXPRESSIONS:
        _produce_test(locals(), gen_test_expression, (expr + "\n", expected))
    for expr, expected in SAMPLE_PY2_EXPRESSIONS:
        _produce_test(locals(), gen_test_expression, (expr + "\n", expected, True))



#!/usr/bin/env python3
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Any, List, Optional

from attr import dataclass
from fissix.fixer_util import LParen, Name

from .helpers import find_last
from .types import (
    ARG_END,
    ARG_LISTS,
    LN,
    STARS,
    SYMBOL,
    TOKEN,
    Capture,
    IMRError,
    Leaf,
    Node,
)

log = logging.getLogger(__name__)


@dataclass
class FunctionArgument:
    name: str = ""
    value: Any = None
    annotation: str = ""
    star: Optional[Leaf] = None
    prefix: Optional[str] = None

    @classmethod
    def build(cls, leaf: Leaf, is_def: bool, **kwargs: Any) -> "FunctionArgument":
        while leaf is not None and leaf.type not in ARG_END:
            if leaf.type in (SYMBOL.star_expr, SYMBOL.argument):
                return cls.build(leaf.children[0], is_def, prefix=leaf.prefix)

            elif leaf.type in STARS:
                kwargs["star"] = leaf.clone()

            elif leaf.type == SYMBOL.tname:
                kwargs["name"] = leaf.children[0].value
                kwargs["annotation"] = leaf.children[-1].value

            elif leaf.type == TOKEN.EQUAL:
                pass

            elif leaf.type == TOKEN.NAME:
                if (is_def and "name" not in kwargs) or (
                    leaf.next_sibling and leaf.next_sibling.type == TOKEN.EQUAL
                ):
                    kwargs["name"] = leaf.value
                else:
                    kwargs["value"] = leaf.clone()

            else:
                # assume everything else is a complex value
                kwargs["value"] = leaf.clone()

            kwargs.setdefault("prefix", leaf.prefix)
            leaf = leaf.next_sibling

        return FunctionArgument(**kwargs)

    @classmethod
    def build_list(
        cls, arguments: List[Leaf], is_def: bool
    ) -> List["FunctionArgument"]:
        result: List[FunctionArgument] = []

        # empty function
        if not arguments:
            return result

        # only care about what's on the inside
        if arguments[0].type in ARG_LISTS:
            leaf = arguments[0].children[0]
        else:
            leaf = arguments[0]

        while leaf is not None:
            arg = cls.build(leaf, is_def)
            log.debug(f"{leaf} -> {arg}")
            result.append(arg)

            # consume leafs for this argument
            while leaf is not None and leaf.type not in ARG_END:
                log.debug(f"consuming {leaf}")
                leaf = leaf.next_sibling

            # assume we stopped on a comma or parenthesis
            if leaf:
                log.debug(f"separator {leaf}")
                leaf = leaf.next_sibling

        return result

    def explode(self, is_def: bool, prefix: str = "") -> List[LN]:
        result: List[LN] = []
        prefix = self.prefix if self.prefix else prefix
        if is_def:
            if self.star:
                self.star.prefix = prefix
                result.append(self.star)
                prefix = ""

            if self.annotation:
                result.append(
                    Node(
                        SYMBOL.tname,
                        [
                            Name(self.name, prefix=prefix),
                            Leaf(TOKEN.COLON, ":", prefix=""),
                            Name(self.annotation, prefix=" "),
                        ],
                        prefix=prefix,
                    )
                )
            else:
                result.append(Name(self.name, prefix=prefix))

            if self.value:
                space = " " if self.annotation else ""
                result.append(Leaf(TOKEN.EQUAL, "=", prefix=space))
                result.append(self.value)

        else:
            if self.star:
                if self.star.type == TOKEN.STAR:
                    node = Node(SYMBOL.star_expr, [self.star], prefix=prefix)
                elif self.star.type == TOKEN.DOUBLESTAR:
                    node = Node(SYMBOL.argument, [self.star], prefix=prefix)

                if self.value:
                    self.value.prefix = ""
                    node.append_child(self.value)

                result.append(node)
                return result

            if self.name:
                self.value.prefix = ""
                result.append(
                    Node(
                        SYMBOL.argument,
                        [
                            Name(self.name, prefix=prefix),
                            Leaf(TOKEN.EQUAL, value="=", prefix=""),
                            self.value,
                        ],
                        prefix=prefix,
                    )
                )
            else:
                self.value.prefix = prefix
                result.append(self.value)

        return result

    @classmethod
    def explode_list(
        cls, arguments: List["FunctionArgument"], is_def: bool
    ) -> Optional[LN]:
        nodes: List[LN] = []
        prefix = ""
        index = 0
        for argument in arguments:
            if index:
                nodes.append(Leaf(TOKEN.COMMA, ",", prefix=""))
                prefix = " "

            result = argument.explode(is_def, prefix=prefix)
            log.debug(f"{argument} -> {result}")
            nodes.extend(result)
            index += 1

        if not nodes:
            return None

        if len(nodes) == 1:
            return nodes[0]

        elif is_def:
            return Node(SYMBOL.typedargslist, nodes, prefix=nodes[0].prefix)

        else:
            return Node(SYMBOL.arglist, nodes, prefix=nodes[0].prefix)


@dataclass
class FunctionSpec:
    name: str
    arguments: List[FunctionArgument]
    is_def: bool
    capture: Capture
    node: Node

    @classmethod
    def build(cls, node: Node, capture: Capture) -> "FunctionSpec":
        try:
            name = capture["function_name"]
            is_def = "function_def" in capture
            args = capture["function_arguments"]
        except KeyError as e:
            raise IMRError("function spec invalid") from e

        arguments = FunctionArgument.build_list(args, is_def)

        return FunctionSpec(name.value, arguments, is_def, capture, node)

    def explode(self) -> LN:
        arguments = FunctionArgument.explode_list(self.arguments, self.is_def)

        rparen = find_last(self.capture["function_parameters"], TOKEN.RPAR)
        rprefix = rparen.prefix if rparen else ""

        if self.is_def:
            parameters = Node(
                SYMBOL.parameters,
                [LParen(), Leaf(TOKEN.RPAR, ")", prefix=rprefix)],
                prefix="",
            )
        else:
            parameters = Node(
                SYMBOL.trailer,
                [LParen(), Leaf(TOKEN.RPAR, ")", prefix=rprefix)],
                prefix="",
            )

        if arguments:
            parameters.insert_child(1, arguments)

        self.capture["function_parameters"].replace(parameters)

        return self.node



#!/usr/bin/env python3
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""Safe code refactoring for modern Python projects."""

__author__ = "John Reese, Facebook"
__version__ = "0.8.0"

from .imr import FunctionArgument, FunctionSpec
from .query import Query
from .tool import BowlerTool
from .types import (
    ARG_ELEMS,
    ARG_END,
    ARG_LISTS,
    DROP,
    LN,
    STARS,
    START,
    SYMBOL,
    TOKEN,
    BowlerException,
    Callback,
    Capture,
    Filename,
    Filter,
    Fixers,
    Hunk,
    IMRError,
    Processor,
    Stringish,
)



#!/usr/bin/env python3
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import functools
import multiprocessing
import sys
import tempfile
import unittest
from contextlib import contextmanager
from io import StringIO

import click
from fissix import pygram, pytree
from fissix.pgen2.driver import Driver

from bowler import Query
from bowler.types import LN, SYMBOL, TOKEN


class BowlerTestCase(unittest.TestCase):
    """Subclass of TestCase that captures stdout and makes it easier to run Bowler."""

    def setUp(self):
        self.buffer = StringIO()
        # Replace the write method instead of stdout so that already-existing
        # loggers end up writing here.
        sys.stdout._saved_write = sys.stdout.write
        sys.stdout.write = self.buffer.write
        sys.stdout._saved_isatty = sys.stdout.isatty
        sys.stdout.isatty = lambda: False

    def tearDown(self):
        if hasattr(sys.stdout, "_saved_write"):
            sys.stdout.write = sys.stdout._saved_write
            del sys.stdout._saved_write
        if hasattr(sys.stdout, "_saved_isatty"):
            sys.stdout.isatty = sys.stdout._saved_isatty
            del sys.stdout._saved_isatty

    def _formatMessage(self, msg1, msg2):
        stdout_text = self.buffer.getvalue()
        msg = msg1 or msg2
        if stdout_text:
            msg += "\n"
            msg += "-" * 20 + "< captured stdout >" + "-" * 20 + "\n"
            msg += stdout_text + "\n"
            msg += "-" * 20 + "<    end stdout   >" + "-" * 20 + "\n"
        return msg

    def run_bowler_modifier(
        self,
        input_text,
        selector=None,
        modifier=None,
        selector_func=None,
        modifier_func=None,
        in_process=True,
        query_func=None,
    ):
        """Returns the modified text."""

        if not (selector or selector_func or query_func):
            raise ValueError("Pass selector")
        if not (modifier or modifier_func or query_func):
            raise ValueError("Pass modifier")

        exception_queue = multiprocessing.Queue()

        def store_exceptions_on(func):
            @functools.wraps(func)
            def inner(node, capture, filename):
                # When in_process=False, this runs in another process.  See notes below.
                try:
                    return func(node, capture, filename)
                except Exception as e:
                    exception_queue.put(e)

            return inner

        def default_query_func(files):
            if selector_func:
                q = selector_func(files)
            else:
                q = Query(files).select(selector)

            if modifier_func:
                q = modifier_func(q)
            else:
                q = q.modify(modifier)

            return q

        if query_func is None:
            query_func = default_query_func

        with tempfile.NamedTemporaryFile(suffix=".py") as f:
            # TODO: I'm almost certain this will not work on Windows, since
            # NamedTemporaryFile has it already open for writing.  Consider
            # using mktemp directly?
            with open(f.name, "w") as fw:
                fw.write(input_text + "\n")

            query = query_func([f.name])
            assert query is not None, "Remember to return the Query"
            assert query.retcode is None, "Return before calling .execute"
            assert len(query.transforms) == 1, "TODO: Support multiple"

            for i in range(len(query.current.callbacks)):
                query.current.callbacks[i] = store_exceptions_on(
                    query.current.callbacks[i]
                )

            # We require the in_process parameter in order to record coverage properly,
            # but it also helps in bubbling exceptions and letting tests read state set
            # by modifiers.
            query.execute(
                interactive=False, write=True, silent=False, in_process=in_process
            )

            # In the case of in_process=False (mirroring normal use of the tool) we use
            # the queue to ship back exceptions from local_process, which can actually
            # fail the test.  Normally exceptions in modifiers are not printed
            # at all unless you pass --debug, and even then you don't get the
            # traceback.
            # See https://github.com/facebookincubator/Bowler/issues/63
            if not exception_queue.empty():
                raise AssertionError from exception_queue.get()

            with open(f.name, "r") as fr:
                return fr.read().rstrip()

    def run_bowler_modifiers(
        self, cases, selector=None, modifier=None, query_func=None
    ):
        for input, expected in cases:
            with self.subTest(input):
                output = self.run_bowler_modifier(
                    input, selector, modifier, query_func=query_func
                )
                self.assertMultiLineEqual(expected, output)

    def parse_line(self, source: str) -> LN:
        grammar = pygram.python_grammar_no_print_statement
        driver = Driver(grammar, convert=pytree.convert)
        # Skip file_input, simple_stmt
        return driver.parse_string(source + "\n").children[0].children[0]


class BowlerTestCaseTest(BowlerTestCase):
    def test_stdout_capture(self):
        print("hi")
        print("there")
        self.assertIn("hi\n", self.buffer.getvalue())

    def test_stdout_click_no_colors(self):
        # This tests that we patched isatty correctly.
        click.echo(click.style("hi", fg="red", bold=True))
        self.assertEqual("hi\n", self.buffer.getvalue())

    def test_run_bowler_modifier(self):
        input = "x=a*b"

        selector = "term< any op='*' any >"

        def modifier(node, capture, filename):
            capture["op"].value = "/"
            capture["op"].changed()

        output = self.run_bowler_modifier(input, selector, modifier)
        self.assertEqual("x=a/b", output)

    def test_run_bowler_modifier_parse_error(self):
        input = "    if x:\n bad"
        selector = "any"
        output = self.run_bowler_modifier(input, selector, lambda *args: None)
        self.assertFalse("None" in output)

    def test_run_bowler_modifier_query_func(self):
        input = "x=a*b"

        selector = "term< any op='*' any >"

        def modifier(node, capture, filename):
            capture["op"].value = "/"
            capture["op"].changed()

        def query_func(arg):
            return Query(arg).select(selector).modify(modifier)

        output = self.run_bowler_modifier(input, query_func=query_func)
        self.assertEqual("x=a/b", output)

    def test_run_bowler_modifier_modifier_func(self):
        input = "x=a*b"

        selector = "term< any op='*' any >"

        def selector_func(arg):
            return Query(arg).select(selector)

        def modifier(node, capture, filename):
            capture["op"].value = "/"
            capture["op"].changed()

        def modifier_func(q):
            return q.modify(modifier)

        output = self.run_bowler_modifier(
            input, selector_func=selector_func, modifier_func=modifier_func
        )
        self.assertEqual("x=a/b", output)

    def test_run_bowler_modifier_ferries_exception(self):
        input = "x=a*b"

        selector = "term< any op='*' any >"

        def modifier(not_enough_args):
            pass

        # Should work in both modes
        self.assertRaises(
            AssertionError,
            lambda: self.run_bowler_modifier(
                input, selector, modifier, in_process=False
            ),
        )

        self.assertRaises(
            AssertionError,
            lambda: self.run_bowler_modifier(
                input, selector, modifier, in_process=True
            ),
        )

    def test_parse_line_leaf(self):
        input = "2.5"
        tree = self.parse_line(input)
        self.assertEqual(TOKEN.NUMBER, tree.type)
        self.assertEqual("2.5", tree.value)

    def test_parse_line_node(self):
        input = "x = (y+1)"
        tree = self.parse_line(input)
        self.assertEqual(SYMBOL.expr_stmt, tree.type)

        self.assertEqual(TOKEN.NAME, tree.children[0].type)
        self.assertEqual(TOKEN.EQUAL, tree.children[1].type)
        self.assertEqual(SYMBOL.atom, tree.children[2].type)

        self.assertEqual("x", tree.children[0].value)



from io import StringIO

from fissix.pgen2 import driver
from fissix import pytree
from fissix.pygram import python_grammar, python_symbols

def code_repr(src: str):
    """
    convert source code to node tree, e.g.
    src = 'path.to.api'
    output = "Node(power, [Leaf(1, 'path'), Node(trailer, [Leaf(23, '.'), Leaf(1, 'to')]), Node(trailer, [Leaf(23, '.'), Leaf(1, 'api')])])"
    """
    driver_ = driver.Driver(python_grammar, convert=pytree.convert)
    tree = driver_.parse_stream(StringIO(src + '\n'))
    ret = tree.children[0].children[0]
    ret.parent = None
    return ret

def replace_module_path(node, before, after):
    """
    replace before with after in node.children, e.g.
    before = 'path1.to1.api1'
    after = 'path2.to2.api2'

    before replace, node is:
    Node(power, [Leaf(1, 'path1'), Node(trailer, [Leaf(23, '.'), Leaf(1, 'to1')]), Node(trailer, [Leaf(23, '.'), Leaf(1, 'api1')])])

    after replace, node is:
    Node(power, [Leaf(1, 'path2'), Node(trailer, [Leaf(23, '.'), Leaf(1, 'to2')]), Node(trailer, [Leaf(23, '.'), Leaf(1, 'api2')])])

    """
    before_parts = before.split('.')
    after_node = code_repr(after)
    # reserve prefix
    after_node.children[0].prefix = node.children[0].prefix
    del node.children[:len(before_parts)]
    for i in range(len(after_node.children)):
        node.insert_child(i, after_node.children[i])



#!/usr/bin/env python3
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""bowler.type_inference

Given an expression, find its result type.

For sufficiently obvious expressions, we can find this using only local
knowledge (numeric literals, and functions/names which have a standard
meaning).  Some obvious examples:

`1.0` -> InferredType.FLOAT
`2L` -> InferredType.INT
`1/2` -> Depends on use_py2_division

Even in cases where we don't know the full inputs, we can make some reasonable
assumptions.  For example when passing `use_py2_division=False,
type_for_unknown=InferredType.INT_OR_FLOAT`:

`x+1.0` -> InferredType.FLOAT
`x/y` -> InferredType.FLOAT
`len(x) + 1` -> InferredType.INT
`float(z)` -> InferredType.FLOAT

This is intended to be useful for either refactoring or flagging for humans
syntax like `range(float(...))` or `"%d" % (float(...),)`.
"""

import enum
from typing import Dict, Sequence, Union

from fissix import pygram, pytree
from fissix.pgen2 import token
from fissix.pgen2.driver import Driver

from .helpers import is_call_to
from .types import LN, SYMBOL, TOKEN

__all__ = ["InferredType", "numeric_expr_type"]


class InferredType(enum.IntEnum):
    # The order of these is important, such that an expression using py3
    # semantics most operators take max(OP_MIN_TYPE[op], max(children)) as the
    # result.
    UNSET = 0
    BOOL = 1
    INT = 2
    # This represents UNKNOWN but assumed to be restricted to normal numeric
    # values.  It can still be promoted to COMPLEX or FLOAT, but if is the
    # final result should be treated as INT (or better).
    INT_OR_FLOAT = 3
    FLOAT = 4
    COMPLEX = 5
    UNKNOWN = 6


# Note: SLASH and DOUBLESLASH are specialcased.
OP_MIN_TYPE: Dict = {
    TOKEN.PLUS: InferredType.INT,
    TOKEN.MINUS: InferredType.INT,
    TOKEN.STAR: InferredType.INT,
    TOKEN.PERCENT: InferredType.INT,
    TOKEN.SLASH: InferredType.INT,
    TOKEN.DOUBLESLASH: InferredType.INT,
    TOKEN.TILDE: InferredType.INT,  # bitwise not
    TOKEN.DOUBLESTAR: InferredType.INT,
    TOKEN.LEFTSHIFT: InferredType.INT,
    TOKEN.RIGHTSHIFT: InferredType.INT,
    TOKEN.VBAR: InferredType.BOOL,
    TOKEN.CIRCUMFLEX: InferredType.BOOL,
    TOKEN.AMPER: InferredType.BOOL,
    TOKEN.LESS: InferredType.BOOL,
}


def numeric_expr_type(
    node: LN,
    use_py2_division=False,
    type_for_unknown: InferredType = InferredType.UNKNOWN,
) -> "InferredType":
    """Infer the type of an expression from its literals.

    We broaden the definition of "literal" a bit to also include calls to
    certain functions like int() and float() where the return type does not
    change based on the arguments.

    Args:
        node: A Node or leaf.
        use_py2_division: Whether to use magical python 2 style division.
        type_for_unknown: An InferredType to customize how you wan unknown
            handled.  Use `INT_OR_FLOAT` if you trust your input to only work
            on numbers, but `UNKNOWN` if you want objects to be an option.

    Returns: InferredType
    """
    if node.type == TOKEN.NUMBER:
        # It's important that we do not use eval here; some literals like `2L`
        # may be invalid in the current interpreter.
        if "j" in node.value:
            return InferredType.COMPLEX
        elif "." in node.value or "e" in node.value:
            return InferredType.FLOAT
        return InferredType.INT
    elif node.type == TOKEN.NAME and node.value in ("True", "False"):
        return InferredType.BOOL
    # TODO let the caller provide other known return types, or even a
    # collection of locals and their types.
    elif is_call_to(node, "bool"):
        return InferredType.BOOL
    elif is_call_to(node, "int") or is_call_to(node, "len"):
        return InferredType.INT
    elif is_call_to(node, "float"):
        return InferredType.FLOAT

    elif node.type in (SYMBOL.comparison, SYMBOL.not_test):
        return InferredType.BOOL
    elif node.type == SYMBOL.factor:
        # unary ~ + -, always [op, number]
        return max(
            OP_MIN_TYPE[node.children[0].type],
            numeric_expr_type(node.children[1], use_py2_division, type_for_unknown),
        )
    elif node.type == SYMBOL.shift_expr:
        # << only valid on int
        return InferredType.INT

    elif node.type == SYMBOL.power:
        # a**b, but also f(...)
        if node.children[1].type != TOKEN.DOUBLESTAR:
            # probably f(...)
            return type_for_unknown

        return max(
            max(OP_MIN_TYPE[c.type] for c in node.children[1::2]),
            max(
                numeric_expr_type(c, use_py2_division, type_for_unknown)
                for c in node.children[::2]
            ),
        )
    elif node.type in (
        SYMBOL.arith_expr,
        SYMBOL.xor_expr,
        SYMBOL.and_expr,
        SYMBOL.expr,
    ):
        return max(
            max(OP_MIN_TYPE[c.type] for c in node.children[1::2]),
            max(
                numeric_expr_type(c, use_py2_division, type_for_unknown)
                for c in node.children[::2]
            ),
        )
    elif node.type == SYMBOL.term:
        # */%
        # This is where things get interesting, as we handle use_py2_division.
        t = InferredType.UNSET
        last_op = None
        for i in range(len(node.children)):
            if i % 2 == 0:
                new = numeric_expr_type(
                    node.children[i], use_py2_division, type_for_unknown
                )
                if last_op == TOKEN.DOUBLESLASH:
                    t = InferredType.INT
                elif last_op == TOKEN.SLASH:
                    if use_py2_division:
                        if t == InferredType.INT and new == InferredType.INT:
                            t = InferredType.INT
                        else:
                            t = max(t, max(OP_MIN_TYPE[last_op], new))
                    else:
                        t = max(t, InferredType.FLOAT)
                else:
                    if last_op:
                        t = max(t, OP_MIN_TYPE[last_op])
                    t = max(t, new)
            else:
                last_op = node.children[i].type
        return t
    elif node.type in (SYMBOL.or_test, SYMBOL.and_test):
        return max(
            numeric_expr_type(c, use_py2_division, type_for_unknown)
            for c in node.children[::2]
        )

    return type_for_unknown



from bowler import Query
from bowler.helpers import power_parts, quoted_parts, dotted_parts
from bowler.types import LN, Capture, Filename, SYMBOL, TOKEN

from fissix.pytree import Leaf, Node, type_repr
from fissix.fixer_util import Attr, Comma, Dot, LParen, Name, Newline, RParen, KeywordArg
from fissix.fixer_util import is_import, touch_import, find_root
from fissix.pygram import python_grammar, python_symbols
from fissix.patcomp import PatternCompiler

from paddle1to2.common import logger
from paddle1to2 import processors, fixers, utils


# don't change the order if you don't know what you are doing.
__all__ = [
    'refactor_demo',
    'refactor_import',
    'norm_api_alias',
    'args_to_kwargs',
    'args_warning',
    'refactor_kwargs',
    'api_rename_and_warning',
    'refactor_syntax',
    'post_refactor',
    ]

def refactor_demo(q: Query, change_spec) -> "Query":
    #q.select_function("old_api").is_call().rename("new_api").process(processors.demo_post_processor)

    #q.fixer(fixers.FixerDemo)
    return q

def refactor_import(q: Query, change_spec) -> "Query":
    """
    1. add "import paddle" if needed.
    2. remove "import paddle.mod" if needed.
    3. remove "import paddle.module as mod", and convert "mod.api" to "paddle.module.api"
    4. remove "from paddle.module import api", and convert "api" to "paddle.module.api"
    """

    # select import_name and import_from
    pattern = """
        (
            file_input< any* >
         |
            name_import=import_name< 'import' '{name}' >
         |
            as_import=import_name< 'import'
                (
                    module_name='{name}'
                |
                    module_name=dotted_name< {dotted_name} any* >
                |
                    dotted_as_name<
                        (
                            module_name='{name}'
                        |
                            module_name=dotted_name< {dotted_name} any* >
                        )
                        'as' module_nickname=any
                    >
                )
            >
        |
            from_import=import_from< 'from'
                (
                    module_name='{name}'
                |
                    module_name=dotted_name< {dotted_name} any* >
                )
                'import' ['(']
                (
                    import_as_name<
                        module_import=any
                        'as'
                        module_nickname=any
                    >*
                |
                    import_as_names<
                        module_imports=any*
                    >
                |
                    module_import=any
                )
             [')'] >
        |
             leaf_node=NAME
        )
    """
    _kwargs = {}
    _kwargs['name'] = 'paddle'
    _kwargs["dotted_name"] = " ".join(quoted_parts(_kwargs["name"]))
    _kwargs["power_name"] = " ".join(power_parts(_kwargs["name"]))
    pattern = pattern.format(**_kwargs)

    imports_map = {}
    paddle_imported = False
    paddle_found = False

    def _find_imports(node: LN, capture: Capture, filename: Filename) -> bool:
        nonlocal paddle_imported, paddle_found
        if not is_import(node):
            return True
        if capture and 'name_import' in capture:
            paddle_imported = True
            paddle_found = True
        if capture and ('module_import' in capture or 'module_imports' in capture or 'module_nickname' in capture):
            paddle_found = True
            if filename not in imports_map:
                imports_map[filename] = {}
            if 'module_import' in capture:
                leaf = capture['module_import']
                if leaf.type == TOKEN.NAME:
                    old_name = leaf.value.strip()
                    new_name = str(capture['module_name']).strip() + '.' + old_name
                    imports_map[filename][old_name] = new_name
            if 'module_imports' in capture:
                for leaf in capture['module_imports']:
                    if leaf.type == TOKEN.NAME:
                        old_name = leaf.value.strip()
                        new_name = str(capture['module_name']).strip() + '.' + old_name
                        imports_map[filename][old_name] = new_name
            if 'module_nickname' in capture:
                old_name = str(capture['module_nickname']).strip()
                new_name = str(capture['module_name']).strip()
                imports_map[filename][old_name] = new_name
        return True

    q.select(pattern).filter(_find_imports)
    # convert to full module path
    def _full_module_path(node: LN, capture: Capture, filename: Filename) -> None:
        if not (isinstance(node, Leaf) and node.type == TOKEN.NAME):
            return
        if filename not in imports_map:
            return
        logger.debug(f"{filename} [{list(capture)}]: {node}")

        # skip import statement
        p = node.parent
        while p is not None:
            if p.type in {SYMBOL.import_name, SYMBOL.import_from}:
                return
            p = p.parent
        # skip if it's already a full module path
        if node.prev_sibling is not None and node.prev_sibling.type == TOKEN.DOT:
            return

        rename_dict = imports_map[filename]
        if node.value in rename_dict:
            # find old_name and new_name
            old_name = node.value
            new_name = rename_dict[old_name]
            if node.parent is not None:
                new_node = utils.code_repr(new_name)
                new_node.children[0].prefix = node.prefix
                if node.parent.type == SYMBOL.power:
                    node.replace(new_node.children)
                else:
                    node.replace(new_node)
    q.modify(_full_module_path)

    # remove as_import and from_import
    def _remove_import(node: LN, capture: Capture, filename: Filename) -> None:
        if not is_import(node):
            return
        _node = capture.get('as_import', None) or capture.get('from_import', None)
        if _node is not None:
            prefix = _node.prefix
            p = _node.parent
            _node.remove()
            # delete NEWLINE node after delete as_import or from_import
            if p and p.children and len(p.children) == 1 and p.children[0].type == TOKEN.NEWLINE:
                p.children[0].remove()
                # restore comment
                p.next_sibling.prefix = prefix + p.next_sibling.prefix
    q.modify(_remove_import)

    # add "import paddle" if needed
    def _add_import(node: LN, capture: Capture, filename: Filename) -> None:
        nonlocal paddle_imported, paddle_found
        if node.type != SYMBOL.file_input:
            return
        if paddle_imported:
            return
        if paddle_found:
            touch_import(None, 'paddle', node)
            logger.info("{}:{} {}".format(filename, node.get_lineno(), 'add "import paddle"'))
            paddle_imported = True
    q.modify(_add_import)

    return q

def norm_api_alias(q: Query, change_spec) -> "Query":
    """
    rename all alias to main alias. e.g.
    origin code snippet:
       ```
       a = path1.to1.alias1()
       ```
    refactored code snippet:
       ```
       a = path2.to2.main_alias()
       ```
    """
    # construct alias mapping
    alias_map = {}
    for main_alias, v in change_spec.items():
        for alias in v.get('alias', []):
            alias_map[alias] = main_alias

    pattern = """ power< 'paddle' trailer< any* >* > """
    def _norm(node: LN, capture: Capture, filename: Filename) -> None:
        code = ''
        for leaf in node.leaves():
            code = code + leaf.value
        found_alias = False
        alias = None
        for _alias in alias_map.keys():
            if code.startswith(_alias):
                found_alias = True
                alias = _alias
                break
        if not found_alias:
            return
        main_alias = alias_map[alias]
        update_to = change_spec[main_alias].get('update_to', None)
        # if main_alias contains "update_to" field, rename alias to "update_to" directly
        if update_to is None:
            utils.replace_module_path(node, alias, main_alias)
        else:
            utils.replace_module_path(node, alias, update_to)
    q.select(pattern).modify(_norm)

    return q

def args_to_kwargs(q:Query, change_spec) -> "Query":
    """
    convert args to kwargs. e.g.
    origin code snippet:
        ```
        a = path.to.api(1, 2)
        ```
    refactored code snippet:
        ```
        a = path.to.api(x=1, y=2)
        ```
    """
    pattern = """
    (
        power< name=any* trailer<  '(' arglist=any* ')' > >
    )
    """
    def _get_func_name(lns: list):
        func_name = ""
        for ln in lns:
            if isinstance(ln, Leaf):
                func_name = func_name + ln.value
            elif isinstance(ln, Node):
                for l in ln.leaves():
                    func_name = func_name + l.value
        return func_name

    def _modify_args_to_kwargs(node, capture, fn):
        args = capture["arglist"]
        name = capture["name"]
        func_name = _get_func_name(name)

        if func_name not in change_spec or 'args_list' not in change_spec[func_name]:
            return

        args_list = change_spec[func_name]['args_list']
        if len(args_list) == 0:
            return

        if isinstance(args[0], Leaf):
            if 1 != len(args_list):
                logger.warn("argument list length not equal, raw func argument list length is 1, but expected length is {}".format(len(arg_list)))
                return

            args[0].replace(KeywordArg(Name(args_list[0]), args[0].clone()))

        elif isinstance(args[0], Node):
            if args[0].type == SYMBOL.arglist:
                print(args[0])
                if len(args[0].children) != (len(args_list) *2-1):
                    logger.warn("argument list length not equal, raw func argument list length is {}, but expected length is {}".format(int((len(args[0].children) +1 )/2), len(args_list)))
                    return

                child = args[0].children
                index = 0
                for ln in child:
                    if ln.type == SYMBOL.argument:
                        index = index + 1
                    elif ln.type != TOKEN.COMMA:
                        ln.replace(KeywordArg(Name(args_list[index]), ln.clone()))
                        index = index + 1
            elif args[0].type == SYMBOL.argument:
                if 1 != len(args_list):
                    logger.warn("argument list length not equal, raw func argument list length is 1, but expected length is {}".format(len(args_list)))
                    return

                raw_arg_name = args[0].children[0].value
                if raw_arg_name != args_list[0]:
                    logger.warn("exist function argument name ({}) not equal expected argument name ({})".format(raw_arg_name, args_list[0]))
                    return


    q.select(pattern).modify(_modify_args_to_kwargs)

    return q

def args_warning(q:Query, change_spec) -> "Query":
    """
    print warning if specified args are used.
    """
    pattern = """
    (
        power< name=any* trailer<  '(' arglist=any* ')' > >
    )
    """
    def _get_func_name(lns: list):
        func_name = ""
        for ln in lns:
            if isinstance(ln, Leaf):
                func_name = func_name + ln.value
            elif isinstance(ln, Node):
                for l in ln.leaves():
                    func_name = func_name + l.value
        return func_name

    def _add_warning(node, capture, fn):
        args = capture["arglist"]
        name = capture["name"]
        func_name = _get_func_name(name)

        if func_name not in change_spec or not change_spec[func_name]["args_warning"]:
            return
        
        args_warning = change_spec[func_name]["args_warning"]

        if args and args[0].type == SYMBOL.arglist:
            child = args[0].children
            for n in child:
                if isinstance(n, Node) and n.type == SYMBOL.argument:
                    arg_name = n.children[0].value
                    if arg_name in args_warning:
                        warning_info = args_warning[arg_name]
                        logger.warn(warning_info)

    q.select(pattern).modify(_add_warning)
    return q

def refactor_kwargs(q:Query, change_spec) -> "Query":
    """
    rename, remove or add kwargs. e.g.
    origin code snippet:
        ```
        a = path.to.api(k1='v1', k2='v2')
        ```
    refactor rule is: [('k1', 'k2_rename'), ('k2', ''), ('', 'k3', 'v3')]
    refactored code snippet:
        ```
        a = path.to.api(k1_rename='v1', k3='v3')
        ```
    """
    return q

def api_rename_and_warning(q:Query, change_spec) -> "Query":
    """
    1. rename old api to new api. e.g.
        origin code snippet:
            ```
            a = old_path.old_to.old_api(1, 2)
            ```
        refactored code snippet:
           ```
           a = new_path.new_to.new_api(1, 2)
           ```
    2. print warning if specified api are used.
    """
    # construct api rename mapping and api warning mapping
    rename_map = {}
    warning_map = {}
    for main_alias, v in change_spec.items():
        new_api_name = v.get('update_to', None)
        if new_api_name is not None:
            rename_map[main_alias] = new_api_name
        warning = v.get('warning', None)
        if warning is not None:
            warning_map[main_alias] = warning

    pattern = """ power< 'paddle' trailer< any* >* > """
    def _rename_and_warning(node: LN, capture: Capture, filename: Filename) -> None:
        code = ''
        for leaf in node.leaves():
            code = code + leaf.value
        found_rename = False
        found_warning = False
        api = None
        for _api in rename_map.keys():
            if code.startswith(_api):
                found_rename = True
                api = _api
                break
        for _api in warning_map.keys():
            if code.startswith(_api):
                found_warning = True
                api = _api
                break
        if not found_rename and not found_warning:
            return
        # if found rename, replace old_api with new_api
        if found_rename:
            utils.replace_module_path(node, api, rename_map[api])
        # if not found rename and found warning, print warning
        elif found_warning:
            warning_msg = "{}:{} {}".format(filename, node.get_lineno(), warning_map[api])
            logger.warning(warning_msg)
    q.select(pattern).modify(_rename_and_warning)

    return q

def refactor_syntax(q:Query, change_spec) -> "Query":
    """
    refactor syntax, such as removing "with" statement. e.g.
    origin code snippet:
        ```
        with paddle.fluid.dygraph.guard(place):
            path.to.api()
        ```
    refactored code snippet:
        ```
        paddle.disable_static(place)
        path.to.api()
        ```
    """
    return q

def post_refactor(q:Query, change_spec) -> "Query":
    """
    post refactor after all prior refactor steps.
    """
    return q




#!/usr/bin/env python3
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

from fissix.pytree import Leaf, Node

from ..helpers import (
    dotted_parts,
    filename_endswith,
    power_parts,
    print_selector_pattern,
    print_tree,
)
from .lib import BowlerTestCase


class PrintTreeTest(BowlerTestCase):
    def test_print_tree_node(self):
        node = self.parse_line("x + 1")
        expected = """\
[arith_expr] ''
.  [NAME] '' 'x'
.  [PLUS] ' ' '+'
.  [NUMBER] ' ' '1'
"""
        print_tree(node)
        self.assertMultiLineEqual(expected, self.buffer.getvalue())

    def test_print_tree_captures(self):
        node = self.parse_line("x + 1 + 2")
        expected = """\
[arith_expr] ''
.  [NAME] '' 'x'
.  [PLUS] ' ' '+'
.  [NUMBER] ' ' '1'
.  [PLUS] ' ' '+'
.  [NUMBER] ' ' '2'
results['op'] =
.  [PLUS] ' ' '+'
results['rhs'] = [Leaf(2, '1'), Leaf(14, '+'), Leaf(2, '2')]
"""
        print_tree(node, {"op": node.children[1], "rhs": node.children[2:]})
        self.assertMultiLineEqual(expected, self.buffer.getvalue())

    def test_print_tree_recurse_limit(self):
        node = self.parse_line("(((x+1)+2)+3)")
        expected = """\
[atom] ''
.  [LPAR] '' '('
.  [arith_expr] ''
.  .  [atom] ''
.  .  .  [LPAR] '' '('
.  .  .  [arith_expr] ''
.  .  .  .  ...
.  .  .  [RPAR] '' ')'
.  .  [PLUS] '' '+'
.  .  [NUMBER] '' '3'
.  [RPAR] '' ')'
"""
        print_tree(node, recurse=3)
        self.assertMultiLineEqual(expected, self.buffer.getvalue())


class PrintSelectorPatternTest(BowlerTestCase):
    def test_print_selector_pattern(self):
        node = self.parse_line("x + 1")
        expected = """\
arith_expr < 'x' '+' '1' > """
        print_selector_pattern(node)
        self.assertMultiLineEqual(expected, self.buffer.getvalue())

    def test_print_selector_pattern_capture(self):
        node = self.parse_line("x + 1")
        expected = """\
arith_expr < 'x' op='+' '1' > """
        print_selector_pattern(node, {"op": node.children[1]})
        self.assertMultiLineEqual(expected, self.buffer.getvalue())

    def test_print_selector_pattern_capture_list(self):
        node = self.parse_line("x + 1")
        # This is not ideal, but hard to infer a good pattern
        expected = """\
arith_expr < 'x' rest='+' rest='1' > """
        print_selector_pattern(node, {"rest": node.children[1:]})
        self.assertMultiLineEqual(expected, self.buffer.getvalue())


class PowerPartsTest(unittest.TestCase):
    def test_power_parts_include_trailer(self):
        self.assertEqual(["'Model'"], power_parts("Model"))
        self.assertEqual(
            ["'models'", "trailer<", "'.'", "'Model'", ">"], power_parts("models.Model")
        )


class DottedPartsTest(unittest.TestCase):
    def test_dotted_parts(self):
        self.assertEqual(["Model"], dotted_parts("Model"))
        self.assertEqual(["models", ".", "Model"], dotted_parts("models.Model"))
        self.assertEqual(
            ["models", ".", "utils", ".", "Model"], dotted_parts("models.utils.Model")
        )


class FilenameEndswithTest(unittest.TestCase):
    def test_single_string(self):
        py = filename_endswith(".py")
        self.assertTrue(py("foo.py"))
        self.assertTrue(py("foo/foo.py"))
        self.assertFalse(py("foo.txt"))
        self.assertFalse(py("foo/foo.txt"))

    def test_sequence(self):
        py = filename_endswith([".py", ".pyi"])
        self.assertTrue(py("foo.py"))
        self.assertTrue(py("foo/foo.py"))
        self.assertTrue(py("foo.pyi"))
        self.assertTrue(py("foo/foo.pyi"))
        self.assertFalse(py("foo.txt"))
        self.assertFalse(py("foo/foo.txt"))



#!/usr/bin/env python3
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import io
import logging
from pathlib import Path
from unittest import TestCase
from unittest.mock import Mock

from fissix.fixer_util import Call, Name

from ..query import Query
from ..types import TOKEN, BadTransform

STDERR = io.StringIO()


class SmokeTest(TestCase):
    @classmethod
    def setUpClass(cls):
        logging.basicConfig(stream=STDERR)

    def test_tr(self):
        target = Path(__file__).parent / "smoke-target.py"

        def takes_string_literal(node, capture, fn):
            args = capture.get("args")
            return args and args[0].type == TOKEN.STRING

        def wrap_string(node, capture, fn):
            args = capture["args"]
            if args and args[0].type == TOKEN.STRING:
                literal = args[0]
                literal.replace(Call(Name("tr"), [literal.clone()]))

        files_processed = []
        hunks_processed = 0

        def verify_hunk(filename, hunk):
            nonlocal hunks_processed
            files_processed.append(filename)
            hunks_processed += 1

            self.assertIn("""-print("Hello world!")""", hunk)
            self.assertIn("""+print(tr("Hello world!"))""", hunk)
            self.assertIn("""-def foo():""", hunk)
            self.assertIn("""+def foo(bar="something"):""", hunk)

        (
            Query(target)
            .select(
                """
                power< "print" trailer< "(" args=any* ")" > >
                """
            )
            .filter(takes_string_literal)
            .modify(wrap_string)
            .select_function("foo")
            .add_argument("bar", '"something"')
            .process(verify_hunk)
            .silent()
        )

        self.assertEqual(hunks_processed, 1)
        self.assertEqual(len(files_processed), 1)
        self.assertIn("smoke-target.py", files_processed[0])

    def test_check_ast(self):
        target = Path(__file__).parent / "smoke-target.py"
        mock_processor = Mock()

        query = (
            Query(str(target))
            .select_function("foo")
            .rename("foo/")
            .process(mock_processor)
            .silent()
        )
        self.assertTrue(any(isinstance(e, BadTransform) for e in query.exceptions))
        mock_processor.assert_not_called()



#!/usr/bin/env python3
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import inspect
import logging
import pathlib
import re
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Type, TypeVar, Union, cast

from attr import Factory, dataclass
from fissix.fixer_base import BaseFix
from fissix.fixer_util import Attr, Comma, Dot, LParen, Name, Newline, RParen
from fissix.pytree import Leaf, Node, type_repr

from .helpers import (
    Once,
    dotted_parts,
    find_first,
    find_last,
    find_previous,
    get_class,
    power_parts,
    print_selector_pattern,
    print_tree,
    quoted_parts,
)
from .imr import FunctionArgument, FunctionSpec
from .tool import BowlerTool
from .types import (
    LN,
    SENTINEL,
    START,
    SYMBOL,
    TOKEN,
    BowlerException,
    Callback,
    Capture,
    Filename,
    FilenameMatcher,
    Filter,
    Hunk,
    Processor,
    Stringish,
    Transform,
)

SELECTORS = {}
Q = TypeVar("Q", bound="Query")
QM = Callable[..., Q]

log = logging.getLogger(__name__)


def selector(pattern: str) -> Callable[[QM], QM]:
    def wrapper(fn: QM) -> QM:
        selector = fn.__name__.replace("select_", "").lower()
        SELECTORS[selector] = pattern

        signature = inspect.signature(fn)
        arg_names = list(signature.parameters)[1:]

        @wraps(fn)
        def wrapped(self: Q, *args, **kwargs) -> Q:
            for arg, value in zip(arg_names, args):
                if hasattr(value, "__name__"):
                    kwargs["source"] = value
                    kwargs[arg] = value.__name__
                else:
                    kwargs[arg] = str(value)

            if "name" in kwargs:
                kwargs["dotted_name"] = " ".join(quoted_parts(kwargs["name"]))
                kwargs["power_name"] = " ".join(power_parts(kwargs["name"]))
            self.transforms.append(Transform(selector, kwargs))
            return self

        return wrapped

    return wrapper


class Query:
    def __init__(
        self,
        *paths: Union[str, List[str]],
        filename_matcher: Optional[FilenameMatcher] = None,
    ) -> None:
        self.paths: List[str] = []
        self.transforms: List[Transform] = []
        self.processors: List[Processor] = []
        self.retcode: Optional[int] = None
        self.filename_matcher = filename_matcher
        self.exceptions: List[BowlerException] = []

        for path in paths:
            if isinstance(path, str):
                self.paths.append(path)
            elif isinstance(path, pathlib.Path):
                self.paths.append(str(path))
            else:
                self.paths.extend(path)

        if not self.paths:
            self.paths.append(".")

    @selector(
        """
        file_input< any* >
    """
    )
    def select_root(self) -> "Query":
        ...

    @selector(
        """
        (
            import_name< 'import'
                (
                    module_name='{name}'
                |
                    module_name=dotted_name< {dotted_name} any* >
                |
                    dotted_as_name<
                        (
                            module_name='{name}'
                        |
                            module_name=dotted_name< {dotted_name} any* >
                        )
                        'as' module_nickname=any
                    >
                )
            >
        |
            import_from< 'from'
                (
                    module_name='{name}'
                |
                    module_name=dotted_name< {dotted_name} any* >
                )
                'import' ['(']
                (
                    import_as_name<
                        module_import=any
                        'as'
                        module_nickname=any
                    >*
                |
                    import_as_names<
                        module_imports=any*
                    >
                |
                    module_import=any
                )
             [')'] >
        |
            module_name=power<
                [TOKEN]
                {power_name}
                module_access=trailer< any* >*
            >
        )
    """
    )
    def select_module(self, name: str) -> "Query":
        ...

    @selector(
        """
        (
            class_def=classdef<
                'class' class_name='{name}'
                any*
                suite<
                    any*
                >
                any*
            >
        |
            class_call=power<
                class_name='{name}'
                trailer< '(' class_arguments=any* ')' >
            >
        |
            class_subclass=classdef< any*
                (
                    any* class_name='{name}' any*
                |
                    arglist< any* class_name='{name}' any* >
                )
                any*
                suite<
                    any*
                >
                any*
            >
        |
            class_import=import_from<
                'from' module_name=any
                'import' ['(']
                (
                    import_as_names<
                        any*
                        class_name='{name}'
                        any*
                    >
                |
                    any*
                    class_name='{name}'
                    any*
                )
            [')'] >
        )
    """
    )
    def select_class(self, name: str) -> "Query":
        ...

    @selector(
        """
        (
            class_def=classdef<
                'class' class_name=any '('
                (
                    any* class_ancestor='{name}' any*
                |
                    arglist< any* class_ancestor='{name}' any* >
                )
                any*
                suite<
                    any*
                >
                any*
            >
        )
    """
    )
    def select_subclass(self, name: str) -> "Query":
        ...

    @selector(
        """
        (
            attr_class=classdef< any*
                suite< any*
                    simple_stmt< any*
                        attr_assignment=expr_stmt<
                            attr_name='{name}' attr_value=any*
                        >
                    any* >
                any* >
            any* >
        |
            attr_assignment=expr_stmt<
                power<
                    any*
                    trailer< any* >*
                    trailer< '.'
                        attr_name='{name}'
                    >
                >
                attr_value=any*
            >
        |
            attr_access=power<
                any*
                trailer< any* >*
                trailer< '.'
                    attr_name='{name}'
                any* >
            any* >
        )
    """
    )
    def select_attribute(self, name: str) -> "Query":
        ...

    @selector(
        """
        (
            decorated=decorated<
                decorators=decorators
                function_def=funcdef<
                    'def' function_name='{name}'
                    function_parameters=parameters< '('
                        function_arguments=typedargslist< ( 'self' | 'cls' ) any* >*
                    ')' >
                    any*
                >
            >
        |
            function_def=funcdef<
                'def' function_name='{name}'
                function_parameters=parameters< '('
                    function_arguments=typedargslist< ( 'self' | 'cls' ) any* >*
                ')' >
                any*
            >
        |
            function_call=power<
                any*
                trailer< any* >*
                trailer<
                    '.' function_name='{name}'
                >
                function_parameters=trailer< '(' function_arguments=any* ')' >
                any*
            >
        |
            function_import=import_from<
                'from' module_name=any
                'import' ['(']
                (
                    import_as_names<
                        any*
                        function_name='{name}'
                        any*
                    >
                |
                    any*
                    function_name='{name}'
                    any*
                )
            [')'] >
        )

    """
    )
    def select_method(self, name: str) -> "Query":
        ...

    @selector(
        """
        (
            decorated=decorated<
                decorators=decorators
                function_def=funcdef<
                    'def' function_name='{name}'
                    function_parameters=parameters< '(' function_arguments=any* ')' >
                    any*
                >
            >
        |
            function_def=funcdef<
                'def' function_name='{name}'
                function_parameters=parameters< '(' function_arguments=any* ')' >
                any*
            >
        |
            function_call=power<
                [TOKEN]
                function_name='{name}'
                function_parameters=trailer< '(' function_arguments=any* ')' >
                remainder=any*
            >
        |
            function_import=import_from<
                'from' module_name=any
                'import' ['(']
                (
                    import_as_names<
                        any*
                        function_name='{name}'
                        any*
                    >
                |
                    any*
                    function_name='{name}'
                    any*
                )
            [')'] >
        )
    """
    )
    def select_function(self, name: str) -> "Query":
        ...

    @selector(
        """
        (
            var_assignment=expr_stmt<
                var_name='{name}'
                var_value=any*
            >
        |
            var_name='{name}'
        )
    """
    )
    def select_var(self, name: str) -> "Query":
        ...

    @selector("""{pattern}""")
    def select_pattern(self, pattern: str) -> "Query":
        ...

    def select(self, pattern: str) -> "Query":
        return self.select_pattern(pattern)

    @property
    def current(self) -> Transform:
        if not self.transforms:
            raise ValueError("no selectors used")

        return self.transforms[-1]

    def is_filename(self, include: str = None, exclude: str = None) -> "Query":
        if include:
            regex = re.compile(include)

            def filter_filename_include(
                node: LN, capture: Capture, filename: Filename
            ) -> bool:
                return regex.search(filename) is not None

            self.current.filters.append(filter_filename_include)

        if exclude:
            regex = re.compile(exclude)

            def filter_filename_exclude(
                node: LN, capture: Capture, filename: Filename
            ) -> bool:
                return regex.search(filename) is None

            self.current.filters.append(filter_filename_exclude)

        return self

    def is_call(self) -> "Query":
        def filter_is_call(node: LN, capture: Capture, filename: Filename) -> bool:
            return bool("function_call" in capture or "class_call" in capture)

        self.current.filters.append(filter_is_call)
        return self

    def is_def(self) -> "Query":
        def filter_is_def(node: LN, capture: Capture, filename: Filename) -> bool:
            return bool("function_def" in capture or "class_def" in capture)

        self.current.filters.append(filter_is_def)
        return self

    def in_class(self, class_name: str, include_subclasses: bool = True) -> "Query":
        def filter_in_class(node: LN, capture: Capture, filename: Filename) -> bool:
            while node.parent is not None:
                if node.type == SYMBOL.classdef:
                    if node.children[1].value == class_name:
                        return True
                    if not include_subclasses:
                        return False
                    for leaf in node.leaves():
                        if leaf.type == TOKEN.COLON:
                            break
                        elif leaf.type == TOKEN.NAME and leaf.value == class_name:
                            return True
                    return False
                node = node.parent
            return False

        self.current.filters.append(filter_in_class)
        return self

    def encapsulate(self, internal_name: str = "") -> "Query":
        transform = self.current
        if transform.selector not in ("attribute"):
            raise ValueError("encapsulate requires select_attribute")

        if not any("filter_in_class" in f.__name__ for f in transform.filters):
            raise ValueError("encapsulate requires in_class filter")

        make_property = Once()
        old_name = transform.kwargs["name"]
        new_name = internal_name or f"_{old_name}"

        if new_name.startswith("__"):
            raise ValueError(
                "renaming {old_name} -> {new_name} is dangerous, "
                "please specify internal_name to avoid name mangling"
            )

        def encapsulate_transform(
            node: LN, capture: Capture, filename: Filename
        ) -> None:
            if "attr_assignment" in capture:
                leaf = capture["attr_name"]
                leaf.replace(Name(new_name, prefix=leaf.prefix))

                if make_property:
                    # TODO: capture and use type annotation from original assignment

                    class_node = get_class(node)
                    suite = find_first(class_node, SYMBOL.suite)
                    assert isinstance(suite, Node)
                    indent_node = find_first(suite, TOKEN.INDENT)
                    assert isinstance(indent_node, Leaf)
                    indent = indent_node.value

                    getter = Node(
                        SYMBOL.decorated,
                        [
                            Node(
                                SYMBOL.decorator,
                                [
                                    Leaf(TOKEN.AT, "@"),
                                    Name("property"),
                                    Leaf(TOKEN.NEWLINE, "\n"),
                                ],
                            ),
                            Node(
                                SYMBOL.funcdef,
                                [
                                    Name("def", indent),
                                    Name(old_name, prefix=" "),
                                    Node(
                                        SYMBOL.parameters,
                                        [LParen(), Name("self"), RParen()],
                                    ),
                                    Leaf(TOKEN.COLON, ":"),
                                    Node(
                                        SYMBOL.suite,
                                        [
                                            Newline(),
                                            Leaf(TOKEN.INDENT, indent.value + "    "),
                                            Node(
                                                SYMBOL.simple_stmt,
                                                [
                                                    Node(
                                                        SYMBOL.return_stmt,
                                                        [
                                                            Name("return"),
                                                            Node(
                                                                SYMBOL.power,
                                                                Attr(
                                                                    Name("self"),
                                                                    Name(new_name),
                                                                ),
                                                                prefix=" ",
                                                            ),
                                                        ],
                                                    ),
                                                    Newline(),
                                                ],
                                            ),
                                            Leaf(TOKEN.DEDENT, "\n" + indent),
                                        ],
                                    ),
                                ],
                                prefix=indent,
                            ),
                        ],
                    )

                    setter = Node(
                        SYMBOL.decorated,
                        [
                            Node(
                                SYMBOL.decorator,
                                [
                                    Leaf(TOKEN.AT, "@"),
                                    Node(
                                        SYMBOL.dotted_name,
                                        [Name(old_name), Dot(), Name("setter")],
                                    ),
                                    Leaf(TOKEN.NEWLINE, "\n"),
                                ],
                            ),
                            Node(
                                SYMBOL.funcdef,
                                [
                                    Name("def", indent),
                                    Name(old_name, prefix=" "),
                                    Node(
                                        SYMBOL.parameters,
                                        [
                                            LParen(),
                                            Node(
                                                SYMBOL.typedargslist,
                                                [
                                                    Name("self"),
                                                    Comma(),
                                                    Name("value", prefix=" "),
                                                ],
                                            ),
                                            RParen(),
                                        ],
                                    ),
                                    Leaf(TOKEN.COLON, ":"),
                                    Node(
                                        SYMBOL.suite,
                                        [
                                            Newline(),
                                            Leaf(TOKEN.INDENT, indent + "    "),
                                            Node(
                                                SYMBOL.simple_stmt,
                                                [
                                                    Node(
                                                        SYMBOL.expr_stmt,
                                                        [
                                                            Node(
                                                                SYMBOL.power,
                                                                Attr(
                                                                    Name("self"),
                                                                    Name(new_name),
                                                                ),
                                                            ),
                                                            Leaf(
                                                                TOKEN.EQUAL,
                                                                "=",
                                                                prefix=" ",
                                                            ),
                                                            Name("value", prefix=" "),
                                                        ],
                                                    ),
                                                    Newline(),
                                                ],
                                            ),
                                            Leaf(TOKEN.DEDENT, "\n" + indent),
                                        ],
                                    ),
                                ],
                                prefix=indent,
                            ),
                        ],
                    )

                    suite.insert_child(-1, getter)
                    suite.insert_child(-1, setter)

                    prev = find_previous(getter, TOKEN.DEDENT, recursive=True)
                    curr = find_last(setter, TOKEN.DEDENT, recursive=True)
                    assert isinstance(prev, Leaf) and isinstance(curr, Leaf)
                    prev.prefix, curr.prefix = curr.prefix, prev.prefix
                    prev.value, curr.value = curr.value, prev.value

        transform.callbacks.append(encapsulate_transform)
        return self

    def rename(self, new_name: str) -> "Query":
        transform = self.current
        old_name = transform.kwargs["name"]

        def rename_transform(node: LN, capture: Capture, filename: Filename) -> None:
            log.debug(f"{filename} [{list(capture)}]: {node}")

            # If two keys reference the same underlying object, do not modify it twice
            visited: List[LN] = []
            for _key, value in capture.items():
                log.debug(f"{_key}: {value}")
                if value in visited:
                    continue
                visited.append(value)

                if isinstance(value, Leaf) and value.type == TOKEN.NAME:
                    if value.value == old_name and value.parent is not None:
                        value.replace(Name(new_name, prefix=value.prefix))
                        break
                elif isinstance(value, Node):
                    if type_repr(value.type) == "dotted_name":
                        dp_old = dotted_parts(old_name)
                        dp_new = dotted_parts(new_name)
                        parts = zip(dp_old, dp_new, value.children)
                        for old, new, leaf in parts:
                            if old != leaf.value:
                                break
                            if old != new:
                                leaf.replace(Name(new, prefix=leaf.prefix))

                        if len(dp_new) < len(dp_old):
                            # if new path is shorter, remove excess children
                            del value.children[len(dp_new) : len(dp_old)]
                        elif len(dp_new) > len(dp_old):
                            # if new path is longer, add new children
                            children = [
                                Name(new) for new in dp_new[len(dp_old) : len(dp_new)]
                            ]
                            value.children[len(dp_old) : len(dp_old)] = children

                    elif type_repr(value.type) == "power":
                        # We don't actually need the '.' so just skip it
                        dp_old = old_name.split(".")
                        dp_new = new_name.split(".")

                        for old, new, leaf in zip(dp_old, dp_new, value.children):
                            if isinstance(leaf, Node):
                                name_leaf = leaf.children[1]
                            else:
                                name_leaf = leaf
                            if old != name_leaf.value:
                                break
                            name_leaf.replace(Name(new, prefix=name_leaf.prefix))

                        if len(dp_new) < len(dp_old):
                            # if new path is shorter, remove excess children
                            del value.children[len(dp_new) : len(dp_old)]
                        elif len(dp_new) > len(dp_old):
                            # if new path is longer, add new trailers in the middle
                            for i in range(len(dp_old), len(dp_new)):
                                value.insert_child(
                                    i, Node(SYMBOL.trailer, [Dot(), Name(dp_new[i])])
                                )

        transform.callbacks.append(rename_transform)
        return self

    def add_argument(
        self,
        name: str,
        value: str,
        positional: bool = False,
        after: Stringish = SENTINEL,
        type_annotation: Stringish = SENTINEL,
    ) -> "Query":
        keyword = not positional

        transform = self.current
        if transform.selector not in ("function", "method"):
            raise ValueError("add_argument must follow select_function/select_method")

        # determine correct position (excluding self/cls) to add new argument
        stop_at = -1
        if positional and after not in (SENTINEL, START):
            if "source" not in transform.kwargs:
                raise ValueError(
                    "using after= with positional= requires passing original function"
                )
            signature = inspect.signature(transform.kwargs["source"])
            if after not in (SENTINEL, START) and after not in signature.parameters:
                raise ValueError(f"{after} does not exist in original function")

            names = list(signature.parameters)
            stop_at = names.index(cast(str, after))
            if names[0] in ("self", "cls", "meta"):
                stop_at -= 1

        def add_argument_transform(
            node: Node, capture: Capture, filename: Filename
        ) -> None:
            if "function_def" not in capture and "function_call" not in capture:
                return

            spec = FunctionSpec.build(node, capture)
            done = False
            value_leaf = Name(value)

            if spec.is_def:
                new_arg = FunctionArgument(
                    name,
                    value_leaf if keyword else None,
                    cast(str, type_annotation) if type_annotation != SENTINEL else "",
                )
                for index, argument in enumerate(spec.arguments):
                    if (
                        after == START
                        or after == argument.name
                        or (positional and (argument.value or argument.star))
                        or (
                            keyword
                            and argument.star
                            and argument.star.type == TOKEN.DOUBLESTAR
                        )
                    ):
                        spec.arguments.insert(index, new_arg)
                        done = True
                        break

                if not done:
                    spec.arguments.append(new_arg)

            elif positional:
                new_arg = FunctionArgument(value=value_leaf)
                for index, argument in enumerate(spec.arguments):
                    if argument.star and argument.star.type == TOKEN.STAR:
                        log.debug(f"noping out due to *{argument.name}")
                        done = True
                        break

                    if (
                        after == START
                        or index == stop_at
                        or argument.name
                        or argument.star
                    ):
                        spec.arguments.insert(index, new_arg)
                        done = True
                        break

                if not done:
                    spec.arguments.append(new_arg)

            spec.explode()

        transform.callbacks.append(add_argument_transform)
        return self

    def modify_argument(
        self,
        name: str,
        new_name: Stringish = SENTINEL,
        type_annotation: Stringish = SENTINEL,
        default_value: Stringish = SENTINEL,
    ) -> "Query":
        transform = self.current
        if transform.selector not in ("function", "method"):
            raise ValueError(f"modifier must follow select_function or select_method")

        def modify_argument_transform(
            node: Node, capture: Capture, filename: Filename
        ) -> None:
            if "function_def" not in capture and "function_call" not in capture:
                return

            spec = FunctionSpec.build(node, capture)

            for argument in spec.arguments:
                if argument.name == name:
                    if new_name is not SENTINEL:
                        argument.name = str(new_name)
                    if spec.is_def and type_annotation is not SENTINEL:
                        argument.annotation = str(type_annotation)
                    if spec.is_def and default_value is not SENTINEL:
                        argument.value = Name(default_value, prefix=" ")

            spec.explode()

        transform.callbacks.append(modify_argument_transform)
        return self

    def remove_argument(self, name: str) -> "Query":
        transform = self.current
        if transform.selector not in ("function", "method"):
            raise ValueError(f"modifier must follow select_function or select_method")

        # determine correct position (excluding self/cls) to add new argument
        stop_at = -1
        if "source" not in transform.kwargs:
            raise ValueError("remove_argument requires passing original function")
        signature = inspect.signature(transform.kwargs["source"])
        if name not in signature.parameters:
            raise ValueError(f"{name} does not exist in original function")

        if signature.parameters[name].kind in (
            inspect.Parameter.VAR_KEYWORD,
            inspect.Parameter.VAR_POSITIONAL,
        ):
            raise ValueError("can't remove *args or **kwargs")

        positional = signature.parameters[name].kind in (
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
        )
        names = list(signature.parameters)
        stop_at = names.index(name)
        if names[0] in ("self", "cls", "meta"):
            stop_at -= 1

        def remove_argument_transform(
            node: Node, capture: Capture, filename: Filename
        ) -> None:
            if "function_def" not in capture and "function_call" not in capture:
                return

            spec = FunctionSpec.build(node, capture)

            if spec.is_def or not positional:
                for argument in spec.arguments:
                    if argument.name == name:
                        spec.arguments.remove(argument)
                        break

            else:
                for index, argument in reversed(list(enumerate(spec.arguments))):
                    if argument.name == name:
                        spec.arguments.pop(index)
                        break

                    if index == stop_at and not argument.name and not argument.star:
                        spec.arguments.pop(index)
                        break

            spec.explode()

        transform.callbacks.append(remove_argument_transform)
        return self

    def fixer(self, fx: Type[BaseFix]) -> "Query":
        self.transforms.append(Transform(fixer=fx))
        return self

    def filter(self, callback: Union[str, Callback]) -> "Query":
        if isinstance(callback, str):
            code = compile(callback, "<string>", "eval")

            def callback(node: Node, capture: Capture, filename: Filename) -> bool:
                return bool(eval(code))  # noqa: developer tool

        callback = cast(Callback, callback)
        self.current.filters.append(callback)
        return self

    def modify(self, callback: Union[str, Callback]) -> "Query":
        if isinstance(callback, str):
            code = compile(callback, "<string>", "exec")

            def callback(node: Node, capture: Capture, filename: Filename) -> None:
                exec(code)

        callback = cast(Callback, callback)
        self.current.callbacks.append(callback)
        return self

    def process(self, callback: Processor) -> "Query":
        self.processors.append(callback)
        return self

    def create_fixer(self, transform):
        if transform.fixer:
            bm_compat = transform.fixer.BM_compatible
            pattern = transform.fixer.PATTERN

        else:
            bm_compat = False
            log.debug(f"select {transform.selector}[{transform.kwargs}]")
            pattern = SELECTORS[transform.selector].format(**transform.kwargs)

            pattern = " ".join(
                line
                for wholeline in pattern.splitlines()
                for line in (wholeline.strip(),)
                if line
            )

            log.debug(f"generated pattern: {pattern}")

        filters = transform.filters
        callbacks = transform.callbacks

        log.debug(f"registered {len(filters)} filters: {filters}")
        log.debug(f"registered {len(callbacks)} callbacks: {callbacks}")

        class Fixer(BaseFix):
            PATTERN = pattern  # type: ignore
            BM_compatible = bm_compat

            def transform(self, node: LN, capture: Capture) -> Optional[LN]:
                filename = cast(Filename, self.filename)
                returned_node = None
                if not filters or all(f(node, capture, filename) for f in filters):
                    if transform.fixer:
                        returned_node = transform.fixer().transform(node, capture)
                    for callback in callbacks:
                        if returned_node and returned_node is not node:
                            raise BowlerException(
                                "Only the last fixer/callback may return "
                                "a different node.  See "
                                "https://pybowler.io/docs/api-modifiers"
                            )
                        returned_node = callback(node, capture, filename)
                return returned_node

        return Fixer

    def compile(self) -> List[Type[BaseFix]]:
        if not self.transforms:
            log.debug(f"no selectors chosen, defaulting to select_root")
            self.select_root()

        fixers: List[Type[BaseFix]] = []
        for transform in self.transforms:
            fixers.append(self.create_fixer(transform))

        return fixers

    def execute(self, **kwargs) -> "Query":
        fixers = self.compile()
        if self.processors:

            def processor(filename: Filename, hunk: Hunk) -> bool:
                apply = True
                for p in self.processors:
                    if p(filename, hunk) is False:
                        apply = False
                return apply

            kwargs["hunk_processor"] = processor

        kwargs.setdefault("filename_matcher", self.filename_matcher)
        tool = BowlerTool(fixers, **kwargs)
        self.retcode = tool.run(self.paths)
        self.exceptions = tool.exceptions
        return self

    def dump(self, selector_pattern=False) -> "Query":
        if not selector_pattern:
            for transform in self.transforms:
                transform.callbacks.append(print_tree)
        else:
            for transform in self.transforms:
                transform.callbacks.append(print_selector_pattern)
        return self.execute(write=False)

    def diff(self, interactive: bool = False, **kwargs) -> "Query":
        return self.execute(write=False, interactive=interactive, **kwargs)

    def idiff(self, **kwargs) -> "Query":
        return self.diff(interactive=True, **kwargs)

    def silent(self, **kwargs) -> "Query":
        return self.execute(silent=True, **kwargs)

    def write(self, **kwargs) -> "Query":
        return self.execute(write=True, **kwargs)



from bowler.types import LN, Capture, Filename

__all__ = [
        'print_match',
        ]

def print_match(node: LN, capture: Capture, filename: Filename) -> bool:
    print('filename:', filename)
    print('code:\n"""{}"""\n'.format(str(node)))
    print('capture:\n"""{}"""\n'.format(capture))
    print('-' * 10)
    return True



#!/usr/bin/env python3
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


# todo: i18n
print("Hello world!")


def foo():
    pass



#!/usr/bin/env python3
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .helpers import (
    DottedPartsTest,
    FilenameEndswithTest,
    PowerPartsTest,
    PrintSelectorPatternTest,
    PrintTreeTest,
)
from .lib import BowlerTestCaseTest
from .query import QueryTest
from .smoke import SmokeTest
from .tool import ToolTest
from .type_inference import ExpressionTest, OpMinTypeTest



import os
import unittest
import textwrap
from bowler import Query
from tempfile import NamedTemporaryFile

from refactor import *

def _refactor_helper(refactor_func, input_src, change_spec) -> str:
    try:
        ntf = NamedTemporaryFile(suffix='.py', delete=False)
        ntf.write(input_src.encode('utf-8'))
        ntf.close()
        q = Query(ntf.name)
        refactor_func(q, change_spec=change_spec).execute(interactive=False, write=True, silent=True)
        with open(ntf.name, 'r') as f:
            output_src = f.read()
        return output_src
    finally:
        os.remove(ntf.name)


class TestRefactorDemo(unittest.TestCase):
    change_spec = {}

    def test_rename_1(self):
        input_src = '''
        old_api()
        '''
        expected_src = '''
        new_api()
        '''
        input_src = textwrap.dedent(input_src).strip() + '\n'
        expected_src = textwrap.dedent(expected_src).strip() + '\n'
        output_src = _refactor_helper(refactor_demo, input_src, self.change_spec)
        self.assertEqual(output_src, expected_src)


if __name__ == '__main__':
    unittest.main()



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import sys
import argparse

from bowler import Query

from paddle1to2.common import *
from paddle1to2 import refactor, filters
from paddle1to2.refactor import *
from paddle1to2.spec import change_spec

def should_convert():
    """
    check if convert should be run.
    convert should be interrupted in the following cases:
    1. directory is not a git repo, and there are something not committed.
    2. file has been converted.
    """
    return True

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--log-level", dest="log_level", type=str, choices=["DEBUG", "INFO", "WARNING", "ERROR"], help="Set log level, default is INFO")
    parser.add_argument("--no-log-file", dest="no_log_file", action='store_true', default=False, help="Don't log to file")
    parser.add_argument("--log-filepath", dest="log_filepath", type=str, help='Set log file path, default is "report.log"')
    parser.add_argument("--inpath", required=True, type=str, help='The file or directory path you want to upgrade.')
    parser.add_argument("--write", action='store_true', default=False, help='Modify files in place.')
    parser.add_argument("--refactor", action='append', choices=refactor.__all__, help='This is a debug option. Specify refactor you want to run. If none, all refactors will be run.')
    parser.add_argument("--print-match", action='store_true', default=False, help='This is a debug option. Print matched code and node for each file.')

    args = parser.parse_args()
    if args.refactor:
        args.refactor = set(args.refactor)

    if args.log_level:
        logger.setLevel(args.log_level)
    if not args.no_log_file:
        log_to_file(args.log_filepath)

    if not should_convert():
        logger.error("convert abort!")
        sys.exit(1)

    # refactor code via "Query" step by step.
    q = Query(args.inpath)
    for fn in refactor.__all__:
        refactor_func = getattr(refactor, fn)
        if args.refactor and fn not in args.refactor:
            continue
        assert callable(refactor_func), "{} is not callable.".format(fn)
        logger.debug("run refactor: {}".format(fn))
        if args.print_match:
            refactor_func(q, change_spec).filter(filters.print_match)
        else:
            refactor_func(q, change_spec)

    if args.write:
        # print diff to stdout, and modify file in place.
        q.execute(interactive=False, write=True, silent=False)
        logger.info("refactor finished, and source files are modified")
    else:
        # print diff to stdout
        q.execute(interactive=False, write=False, silent=False)
        logger.info("refactor finished without touching source files")

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from unittest import mock

from ..query import SELECTORS, Query
from ..types import TOKEN, BowlerException, Leaf
from .lib import BowlerTestCase


class QueryTest(BowlerTestCase):
    fake_paths = ["foo/bar", "baz.py"]

    def test_basic(self):
        query = Query(self.fake_paths)
        self.assertEqual(len(query.transforms), 0)
        self.assertEqual(query.paths, self.fake_paths)
        with self.assertRaises(ValueError):
            transform = query.current
            self.assertEqual(transform, None)

        query.select_root().is_filename("frob.py")
        self.assertEqual(len(query.transforms), 1)
        self.assertEqual(query.current.selector, "root")
        self.assertEqual(len(query.current.kwargs), 0)
        self.assertEqual(len(query.current.filters), 1)
        self.assertEqual(len(query.current.filters), 1)

        fixers = query.compile()
        self.assertEqual(len(fixers), 1)
        self.assertEqual(fixers[0].PATTERN, SELECTORS["root"].strip())

    def test_rename_func(self):
        def query_func(arg):
            return Query(arg).select_function("f").rename("foo")

        self.run_bowler_modifiers(
            [
                ("def f(x): pass", "def foo(x): pass"),
                ("def g(x): pass", "def g(x): pass"),
                ("f()", "foo()"),
                ("g()", "g()"),
            ],
            query_func=query_func,
        )

    def test_rename_class(self):
        self.run_bowler_modifiers(
            [("class Bar(Foo):\n  pass", "class FooBar(Foo):\n  pass")],
            query_func=lambda x: Query(x).select_class("Bar").rename("FooBar"),
        )

    def test_rename_module(self):
        self.run_bowler_modifiers(
            [("from a.b.c.d import E", "from a.f.d import E")],
            query_func=lambda x: Query(x).select_module("a.b.c").rename("a.f"),
        )

    def test_rename_module_callsites(self):
        input = """\
a.b.c.d.f()
w.bar()"""

        def selector(pattern):
            def _selector(arg):
                return Query(arg).select_module(pattern)

            return _selector

        def modifier(pattern):
            def _modifier(q):
                return q.rename(pattern)

            return _modifier

        # Same length
        output = self.run_bowler_modifier(
            input, selector_func=selector("a.b.c"), modifier_func=modifier("x.y.z")
        )
        expected = """\
x.y.z.d.f()
w.bar()"""
        self.assertMultiLineEqual(expected, output)

        # Shorter replacement
        output = self.run_bowler_modifier(
            input, selector_func=selector("a.b.c"), modifier_func=modifier("x.y")
        )
        expected = """\
x.y.d.f()
w.bar()"""
        self.assertMultiLineEqual(expected, output)

        # Longer replacement
        output = self.run_bowler_modifier(
            input, selector_func=selector("a.b.c"), modifier_func=modifier("w.x.y.z")
        )
        expected = """\
w.x.y.z.d.f()
w.bar()"""
        self.assertMultiLineEqual(expected, output)

        # Single character replacements replacement
        output = self.run_bowler_modifier(
            input, selector_func=selector("w"), modifier_func=modifier("x.y.z")
        )
        expected = """\
a.b.c.d.f()
x.y.z.bar()"""
        self.assertMultiLineEqual(expected, output)

    def test_rename_subclass(self):
        def query_func(x):
            return Query(x).select_subclass("Foo").rename("somepackage.Foo")

        self.run_bowler_modifiers(
            [("class Bar(Foo):\n  pass", "class Bar(somepackage.Foo):\n  pass")],
            query_func=query_func,
        )

    def test_filter_in_class(self):
        def query_func_bar(x):
            return Query(x).select_function("f").in_class("Bar", False).rename("g")

        def query_func_foo(x):
            return Query(x).select_function("f").in_class("Foo", False).rename("g")

        def query_func_foo_subclasses(x):
            return Query(x).select_function("f").in_class("Foo", True).rename("g")

        # Does not have subclasses enabled
        self.run_bowler_modifiers(
            [
                (
                    "class Bar(Foo):\n  def f(self): pass",
                    "class Bar(Foo):\n  def f(self): pass",
                )
            ],
            query_func=query_func_foo,
        )
        # Does
        self.run_bowler_modifiers(
            [
                (
                    "class Bar(Foo):\n  def f(self2): pass",
                    "class Bar(Foo):\n  def g(self2): pass",
                ),
                (
                    "class Bar(Baz):\n  def f(self2): pass",
                    "class Bar(Baz):\n  def f(self2): pass",
                ),
            ],
            query_func=query_func_foo_subclasses,
        )
        # Operates directly on class
        self.run_bowler_modifiers(
            [
                (
                    "class Bar(Foo):\n  def f(self3): pass",
                    "class Bar(Foo):\n  def g(self3): pass",
                )
            ],
            [("class Bar:\n  def f(self3): pass", "class Bar:\n  def g(self3): pass")],
            query_func=query_func_bar,
        )
        # Works on functions too, not just methods
        self.run_bowler_modifiers(
            [("def f(): pass", "def f(): pass")], query_func=query_func_bar
        )

    def test_add_argument(self):
        def query_func(x):
            return Query(x).select_function("f").add_argument("y", "5")

        self.run_bowler_modifiers(
            [
                ("def f(x): pass", "def f(x, y=5): pass"),
                ("def g(x): pass", "def g(x): pass"),
                # ("f()", "???"),
                ("g()", "g()"),
            ],
            query_func=query_func,
        )

    def test_modifier_return_value(self):
        input = "a+b"

        def modifier(node, capture, filename):
            new_op = Leaf(TOKEN.MINUS, "-")
            return new_op

        output = self.run_bowler_modifier(input, "'+'", modifier)
        self.assertEqual("a-b", output)

    def test_modifier_return_value_multiple(self):
        input = "a+b"

        def noop_modifier(node, capture, filename):
            print("Noop modifier")
            pass

        def modifier(node, capture, filename):
            print("Modifier")
            new_op = Leaf(TOKEN.MINUS, "-")
            return new_op

        def add_ok_modifier(q):
            return q.modify(noop_modifier).modify(modifier)

        output = self.run_bowler_modifier(input, "'+'", modifier_func=add_ok_modifier)
        self.assertEqual("a-b", output)

        def add_bad_modifier(q):
            return q.modify(modifier).modify(noop_modifier)

        with mock.patch("bowler.tool.log.error") as error:
            output = self.run_bowler_modifier(
                input, "'+'", modifier_func=add_bad_modifier
            )
            self.assertEqual("a+b", output)  # unmodified
            self.assertTrue(error.call_args)
            self.assertIn(
                "Only the last fixer/callback may return", error.call_args[0][0]
            )



import logging

from fissix import fixer_base
from fissix.refactor import RefactoringTool

__all__ = [
        'FixerDemo',
        ]

class FixerDemo(fixer_base.BaseFix):
    BM_compatible = True
    # match all function call
    PATTERN = """power< any* >"""

    def __init__(self):
        _logger = logging.getLogger("RefactoringTool")
        super(FixerDemo, self).__init__(RefactoringTool._default_options, _logger)

    def transform(self, node, results):
        print('code passed to transform:', node)
        print('results passed to transform:', results)
        return node


